import numpy as np

from scipy import ndimage

from subpixel_edges.edgepixel import EdgePixel
from subpixel_edges.edges_iter1 import h_edges, v_edges


def main_iter1(F, threshold, iters, order):
    # smooth image
    H = np.ones((3, 3), np.float32) / 9
    w = (1 + 24 * H[1, 2] + 48 * H[1, 1]) / 12
    G = ndimage.convolve(F, H, mode='constant')

    # initialization
    rows, cols = np.shape(G)
    [x, y] = np.meshgrid(np.arange(cols), np.arange(rows))
    ep = EdgePixel()
    max_valid_offset = 1

    # compute partial derivatives
    Gx = np.zeros((rows, cols))
    Gx[0: rows, 1: cols - 1] = 0.5 * (G[0: rows, 2: cols] - G[0: rows, 0: cols - 2])
    Gy = np.zeros((rows, cols))
    Gy[1: rows - 1, 0: cols] = 0.5 * (G[2: rows, 0: cols] - G[0: rows - 2, 0: cols])
    grad = np.sqrt(Gx ** 2 + Gy ** 2)

    # detect edge pixels with maximum Gy (not including margins)
    absGyInner = np.abs(Gy[5:rows - 5, 2: cols - 2])
    absGxInner = np.abs(Gx[2:rows - 2, 5: cols - 5])

    Ey = np.zeros((rows, cols), dtype=np.bool)
    Ex = np.zeros((rows, cols), dtype=np.bool)

    Ey[5: rows - 5, 2: cols - 2] = np.logical_and.reduce([
        grad[5: rows - 5, 2: cols - 2] > threshold,
        absGyInner >= np.abs(Gx[5: rows - 5, 2: cols - 2]),
        absGyInner >= np.abs(Gy[4: rows - 6, 2: cols - 2]),
        absGyInner > np.abs(Gy[6: rows - 4, 2: cols - 2])
    ])

    Ex[2: rows - 2, 5: cols - 5] = np.logical_and.reduce([
        grad[2: rows - 2, 5: cols - 5] > threshold,
        absGxInner > np.abs(Gy[2: rows - 2, 5: cols - 5]),
        absGxInner >= np.abs(Gx[2: rows - 2, 4: cols - 6]),
        absGxInner > np.abs(Gx[2: rows - 2, 6: cols - 4])
    ])

    Ey = Ey.ravel('F')
    Ex = Ex.ravel('F')
    y = y.ravel('F')
    x = x.ravel('F')

    edges_y = (x[Ey] * rows + y[Ey])
    edges_x = (x[Ex] * rows + y[Ex])

    Gx = Gx.ravel('F')
    Gy = Gy.ravel('F')

    FF = F.ravel('F')
    GG = G.ravel('F')

    edges_y, x_y, y_y, nx_y, ny_y, curv_y, i0_y, i1_y = h_edges(x, y, F, G, rows, Gx, Gy, w, max_valid_offset, edges_y,
                                                                order, threshold, FF, GG)
    edges_x, x_x, y_x, nx_x, ny_x, curv_x, i0_x, i1_x = v_edges(x, y, F, G, rows, Gx, Gy, w, max_valid_offset, edges_x,
                                                                order, threshold, FF, GG)

    ep.ny = np.concatenate((ny_y, ny_x), axis=0)
    ep.nx = np.concatenate((nx_y, nx_x), axis=0)

    ep.y = np.concatenate((y_y, y_x), axis=0)
    ep.x = np.concatenate((x_y, x_x), axis=0)

    ep.position = np.concatenate((edges_y, edges_x), axis=0)
    ep.curv = np.concatenate((curv_y, curv_x), axis=0)
    ep.i0 = np.concatenate((i0_y, i0_x), axis=0)
    ep.i1 = np.concatenate((i1_y, i1_x), axis=0)

    return ep



import numpy as np

from subpixel_edges.final_detector_iter0 import main_iter0
from subpixel_edges.final_detector_iter1 import main_iter1
from subpixel_edges.final_detector_iterN import main_iterN


def init():
    """
    Run all the methods to trigger JIT compilation. Only needed the first time
    the code is executed. This method may be removed in future releases, for example if
    AOT compilation is used instead.
    """
    img = np.zeros((8, 8))

    subpixel_edges(img, 15, 0, 2)
    subpixel_edges(img, 15, 1, 2)
    subpixel_edges(img, 15, 2, 2)


def subpixel_edges(img, threshold, iters, order):
    """
    Detects subpixel features for each pixel belonging to an edge in `img`.

    The subpixel edge detection used the method published in the following paper:
    "Accurate Subpixel Edge Location Based on Partial Area Effect"
    http://www.sciencedirect.com/science/article/pii/S0262885612001850

    Parameters
    ----------
    img: ndarray
        A grayscale image.
    threshold: int or float
        Specifies the minimum difference of intensity at both
        sides of a pixel to be considered as an edge.
    iters: int
        Specifies how many smoothing iterations are needed
        to find the final edges:
            0:  Oriented to noise free images. No previous smoothing on
                the image. The detection is applied on the original
                image values (section 3 of the paper).
            1:  Oriented to low-noise images. The detection is applied
                on the image previously smoothed by a 3x3 mask
                (default) (sections 4 and 5 of the paper)
            >1: Oriented to high-noise images. Several stages of
                smoothing + detection + synthetic image creation are
                applied (section 6 of the paper). A few iterations are
                normally enough.
    order: int
        Specifies the order of the edges to find:
            1:  first order edges (straight lines)
            2:  second order edges (default)

    Returns
    -------
    An instance of EdgePixel
    """
    if iters == 0:
        return main_iter0(img, threshold, iters, order)
    elif iters == 1:
        return main_iter1(img, threshold, iters, order)
    elif iters > 1:
        for iterN in range(iters):
            ep, img = main_iterN(img, threshold, iters, order)
        return ep



import numpy as np

from numba import njit


@njit(cache=True)
def h_edges(F, rows, Fx, Fy, edges, order):
    n_edges = np.shape(edges)[0]

    A = np.zeros((n_edges, 1))
    B = np.zeros((n_edges, 1))
    a = np.zeros((n_edges, 1))
    b = np.zeros((n_edges, 1))
    c = np.zeros((n_edges, 1))

    for k in range(n_edges):
        edge = edges[k]

        m1 = -1
        m2 = 1
        if Fx[edge] * Fy[edge] >= 0:
            m = 1
            l1 = 0
            r2 = 0
            min_l1 = -3
            max_r2 = 3
            l2 = 1
            r1 = -1
            max_l2 = 4
            min_r1 = -4
        else:
            m = -1
            l1 = -1
            r2 = 1
            min_l1 = -4
            max_r2 = 4
            l2 = 0
            r1 = 0
            max_l2 = 3
            min_r1 = -3

        if np.abs(Fx[edge]) < 1:
            l1 = -1
            l2 = 1
            r1 = -1
            r2 = 1

        while l1 > min_l1 and np.abs(Fy[edge - rows + l1]) >= np.abs(Fy[edge - rows + l1 - 1]):
            l1 -= 1

        while l2 < max_l2 and np.abs(Fy[edge - rows + l2]) >= np.abs(Fy[edge - rows + l2 + 1]):
            l2 += 1

        while m1 > -4 and np.abs(Fy[edge + m1]) >= np.abs(Fy[edge + m1 - 1]):
            m1 -= 1

        while m2 < 4 and np.abs(Fy[edge + m2]) >= np.abs(Fy[edge + m2 + 1]):
            m2 += 1

        while r1 > min_r1 and np.abs(Fy[edge + rows + r1]) >= np.abs(Fy[edge + rows + r1 - 1]):
            r1 -= 1

        while r2 < max_r2 and np.abs(Fy[edge + rows + r2]) >= np.abs(Fy[edge + rows + r2 + 1]):
            r2 += 1

        if m > 0:
            AA = (F[edge + m2] + F[edge + rows + r2]) / 2
            BB = (F[edge - rows + l1] + F[edge + m1]) / 2
        else:
            AA = (F[edge - rows + l2] + F[edge + m2]) / 2
            BB = (F[edge + m1] + F[edge + rows + r1]) / 2

        SL = 0
        SM = 0
        SR = 0

        for n in range(l1, l2 + 1):
            SL += F[edge - rows + n]

        for n in range(m1, m2 + 1):
            SM += F[edge + n]

        for n in range(r1, r2 + 1):
            SR += F[edge + rows + n]

        den = 2 * (AA - BB)

        if order == 2:
            if den != 0:
                c[k] = (SL + SR - 2 * SM + AA * (2 * m2 - l2 - r2) - BB * (2 * m1 - l1 - r1)) / den

        else:
            c[k] = 0

        if den != 0:
            b[k] = (SR - SL + AA * (l2 - r2) - BB * (l1 - r1)) / den
            a[k] = (2 * SM - AA * (1 + 2 * m2) - BB * (1 - 2 * m1)) / den - c[k] / 12

        A[k] = AA
        B[k] = BB

    return A, B, a, b, c


@njit(cache=True)
def v_edges(F, rows, Fx, Fy, edges, order):
    n_edges = np.shape(edges)[0]

    A = np.zeros((n_edges, 1))
    B = np.zeros((n_edges, 1))
    a = np.zeros((n_edges, 1))
    b = np.zeros((n_edges, 1))
    c = np.zeros((n_edges, 1))

    for k in range(n_edges):
        edge = edges[k]

        m1 = -1
        m2 = 1
        if Fx[edge] * Fy[edge] >= 0:
            m = 1
            l1 = 0
            r2 = 0
            min_l1 = -3
            max_r2 = 3
            l2 = 1
            r1 = -1
            max_l2 = 4
            min_r1 = -4
        else:
            m = -1
            l1 = -1
            r2 = 1
            min_l1 = -4
            max_r2 = 4
            l2 = 0
            r1 = 0
            max_l2 = 3
            min_r1 = -3

        if np.abs(Fy[edge]) < 1:
            l1 = -1
            l2 = 1
            r1 = -1
            r2 = 1

        while l1 > min_l1 and np.abs(Fx[edge - 1 + l1 * rows]) >= np.abs(Fx[edge - 1 + (l1 - 1) * rows]):
            l1 -= 1

        while l2 < max_l2 and np.abs(Fx[edge - 1 + l2 * rows]) >= np.abs(Fx[edge - 1 + (l2 + 1) * rows]):
            l2 += 1

        while m1 > -4 and np.abs(Fx[edge + m1 * rows]) >= np.abs(Fx[edge + (m1 - 1) * rows]):
            m1 -= 1

        while m2 < 4 and np.abs(Fx[edge + m2 * rows]) >= np.abs(Fx[edge + (m2 + 1) * rows]):
            m2 += 1

        while r1 > min_r1 and np.abs(Fx[edge + 1 + r1 * rows]) >= np.abs(Fx[edge + 1 + (r1 - 1) * rows]):
            r1 -= 1

        while r2 < max_r2 and np.abs(Fx[edge + 1 + r2 * rows]) >= np.abs(Fx[edge + 1 + (r2 + 1) * rows]):
            r2 += 1

        if m > 0:
            AA = (F[edge + m2 * rows] + F[edge + 1 + r2 * rows]) / 2
            BB = (F[edge - 1 + l1 * rows] + F[edge + m1 * rows]) / 2
        else:
            AA = (F[edge - 1 + l2 * rows] + F[edge + m2 * rows]) / 2
            BB = (F[edge + m1 * rows] + F[edge + 1 + r1 * rows]) / 2

        SL = 0
        SM = 0
        SR = 0

        for n in range(l1, l2 + 1):
            SL += F[edge - 1 + n * rows]

        for n in range(m1, m2 + 1):
            SM += F[edge + n * rows]

        for n in range(r1, r2 + 1):
            SR += F[edge + 1 + n * rows]

        # compute edge features
        den = 2 * (AA - BB)

        if order == 2:
            if den != 0:
                c[k] = (SL + SR - 2 * SM + AA * (2 * m2 - l2 - r2) - BB * (2 * m1 - l1 - r1)) / den

        else:
            c[k] = 0

        if den != 0:
            b[k] = (SR - SL + AA * (l2 - r2) - BB * (l1 - r1)) / den
            a[k] = (2 * SM - AA * (1 + 2 * m2) - BB * (1 - 2 * m1)) / den - c[k] / 12

        A[k] = AA
        B[k] = BB

    return A, B, a, b, c



import numpy as np

from subpixel_edges.edgepixel import EdgePixel
from subpixel_edges.edges_iter0 import h_edges, v_edges


def main_iter0(F, threshold, iters, order):
    ep = EdgePixel()

    rows, cols = np.shape(F)
    [x, y] = np.meshgrid(np.arange(cols), np.arange(rows))

    Fx = np.zeros((rows, cols))
    Fx[0: rows, 1: cols - 1] = 0.5 * (F[0: rows, 2: cols] - F[0: rows, 0: cols - 2])
    Fy = np.zeros((rows, cols))
    Fy[1: rows - 1, 0: cols] = 0.5 * (F[2: rows, 0: cols] - F[0: rows - 2, 0: cols])
    grad = np.sqrt(Fx ** 2 + Fy ** 2)

    abs_Fy_inner = np.abs(Fy[5:rows - 5, 2: cols - 2])
    abs_Fx_inner = np.abs(Fx[2:rows - 2, 5: cols - 5])

    Ey = np.zeros((rows, cols), dtype=np.bool)
    Ex = np.zeros((rows, cols), dtype=np.bool)

    Ey[5: rows - 5, 2: cols - 2] = np.logical_and.reduce([
        grad[5: rows - 5, 2: cols - 2] > threshold,
        abs_Fy_inner >= np.abs(Fx[5: rows - 5, 2: cols - 2]),
        abs_Fy_inner >= np.abs(Fy[4: rows - 6, 2: cols - 2]),
        abs_Fy_inner > np.abs(Fy[6: rows - 4, 2: cols - 2])
    ])

    Ex[2: rows - 2, 5: cols - 5] = np.logical_and.reduce([
        grad[2: rows - 2, 5: cols - 5] > threshold,
        abs_Fx_inner > np.abs(Fy[2: rows - 2, 5: cols - 5]),
        abs_Fx_inner >= np.abs(Fx[2: rows - 2, 4: cols - 6]),
        abs_Fx_inner > np.abs(Fx[2: rows - 2, 6: cols - 4])
    ])

    Ey = Ey.ravel('F')
    Ex = Ex.ravel('F')
    y = y.ravel('F')
    x = x.ravel('F')

    edges_y = (x[Ey] * rows + y[Ey])
    edges_x = (x[Ex] * rows + y[Ex])

    F = F.ravel('F')
    Fx = Fx.ravel('F')
    Fy = Fy.ravel('F')

    ny = np.ones((np.shape(edges_y)[0], 1))
    ny[Fy[edges_y] < 0] = -1
    nx = np.ones((np.shape(edges_x)[0], 1))
    nx[Fx[edges_x] < 0] = -1

    Ay, By, ay, by, cy = h_edges(F, rows, Fx, Fy, edges_y, order)
    Ax, Bx, ax, bx, cx = v_edges(F, rows, Fx, Fy, edges_x, order)

    ay = ay.ravel('F')
    ax = ax.ravel('F')

    ep.position = np.r_[edges_y, edges_x]
    ep.x = np.r_[x[edges_y], x[edges_x] - ax]
    ep.y = np.r_[y[edges_y] - ay, y[edges_x]]

    ep.nx = np.r_[
        np.sign(Ay - By) / np.sqrt(1 + by ** 2) * by,
        np.sign(Ax - Bx) / np.sqrt(1 + bx ** 2),
    ]
    ep.ny = np.r_[
        np.sign(Ay - By) / np.sqrt(1 + by ** 2),
        np.sign(Ax - Bx) / np.sqrt(1 + bx ** 2) * bx,
    ]
    ep.curv = np.r_[
        2 * cy * ny / ((1 + by ** 2) ** 1.5),
        2 * cx * nx / ((1 + bx ** 2) ** 1.5),
    ]
    ep.i0 = np.r_[np.minimum(Ay, By), np.minimum(Ax, Bx)]
    ep.i1 = np.r_[np.maximum(Ay, By), np.maximum(Ax, Bx)]

    # # erase elements outside the image size
    index_to_erase1 = np.union1d(np.where(ep.x > cols), np.where(ep.y > rows))
    index_to_erase2 = np.union1d(np.where(ep.x < 0), np.where(ep.y < 0))
    
    index_to_erase = np.union1d(index_to_erase1, index_to_erase2)
    
    ep.x = np.delete(ep.x, index_to_erase)
    ep.y = np.delete(ep.y, index_to_erase)
    ep.nx = np.delete(ep.nx, index_to_erase)
    ep.ny = np.delete(ep.ny, index_to_erase)
    ep.i0 = np.delete(ep.i0, index_to_erase)
    ep.i1 = np.delete(ep.i1, index_to_erase)
    ep.position = np.delete(ep.position, index_to_erase)
    ep.curv = np.delete(ep.curv, index_to_erase)

    return ep



import numpy as np
import scipy


class EdgePixel:
    """
    Class describing pixel features.

    position:
        1D-index inside image
    x, y:
        subpixel position
    nx, ny:
        normal vector (normalized)
    curv:
        curvature
    i0, i1:
        intensities at both sides
    """
    def __init__(self, position=None, x=None, y=None, nx=None, ny=None, curv=None, i0=None, i1=None):
        
        self.position = position  # 1D index inside image
        self.x = x                # X subpixel position
        self.y = y                # Y subpixel position
        self.nx = nx              # normal vector (normalized)
        self.ny = ny              # normal vector (normalized)
        self.curv = curv          # curvature
        self.i0 = i0              # intensities
        self.i1 = i1              # intensities

    def save(self, filename, format='npz'):
        if format == 'npz':
            np.savez(filename, **vars(self))
        else:
            raise ValueError('File format "%s" not supported' % format)

    @classmethod
    def load(cls, filename, format='npz'):
        if format == 'npz':
            edges = np.load(filename)

            return EdgePixel(
                position=edges['position'],
                x=edges['x'],
                y=edges['y'],
                nx=edges['nx'],
                ny=edges['ny'],
                curv=edges['curv'],
                i0=edges['i0'],
                i1=edges['i1']
            )
        elif format == 'mat':
            edges = scipy.io.loadmat(filename)

            return EdgePixel(
                position=edges['position'].ravel('F'),
                x=edges['x'].ravel('F'),
                y=edges['y'].ravel('F'),
                nx=edges['nx'].ravel('F'),
                ny=edges['ny'].ravel('F'),
                curv=edges['curv'].ravel('F'),
                i0=edges['i0'].ravel('F'),
                i1=edges['i1'].ravel('F')
            )
        else:
            raise ValueError('File format "%s" not supported' % format)



import setuptools

with open('README.md', 'r') as readme:
    long_description = readme.read()

setuptools.setup(
    name='subpixel-edges',
    version='0.1.0',
    author='GraviToni',
    description='Pure Python implementation of subpixel edge location algorithm based on partial area effect',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/gravi-toni/subpixel-edges',
    license='MIT',
    install_requires=['numpy>=1.18.4',
                      'scipy>=1.4.1',
                      'numba>=0.49'],
    extras_require={
        'tests': ['pytest',
                  'opencv-python>=4.1.0.25'],
        'examples': ['matplotlib>=3.2.1',
                     'opencv-python>=4.1.0.25'],
    },
    packages=setuptools.find_packages(),
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    python_requires='>=3.0',
)



import numpy as np
from numba import njit


@njit(cache=True)
def circle_grid(x, y, radius2, x_center, y_center, dx, dy):
    num_pixels = x.shape[0]

    if num_pixels > 0:
        p = np.zeros((num_pixels, 1))
        for n in range(0, num_pixels):
            grid = ((x[n] + dx - x_center) ** 2 + (y[n] + dy - y_center) ** 2 < radius2) * 1
            p[n] = np.mean(grid)

    else:
        p = np.zeros((0, 0))

    return p


@njit(cache=True)
def circle_horizontal_window(x_window_center, y_window_center, x_center, y_center, radius, inner_intensity,
                             outer_intensity, grid_resolution):
    # compute pixels completely outside or inside
    r2 = radius ** 2

    x = np.zeros(((np.arange(x_window_center - 4, x_window_center + 4 + 1)).size,
                  (np.arange(y_window_center - 1, y_window_center + 1 + 1)).size),
                 dtype=np.float64).T
    y = x.copy()
    x[:, :] = np.arange(x_window_center - 4, x_window_center + 4 + 1)
    y_vect = np.arange(y_window_center - 1, y_window_center + 1 + 1)
    y_vect = np.reshape(y_vect, (y_vect.shape[0], 1))
    y[:, :] = y_vect

    c = ((x - 0.5 - x_center) ** 2 + (y - 0.5 - y_center) ** 2 < r2).astype(np.float64)
    c = c + ((x - 0.5 - x_center) ** 2 + (y + 0.5 - y_center) ** 2 < r2)
    c = c + ((x + 0.5 - x_center) ** 2 + (y - 0.5 - y_center) ** 2 < r2)
    c = c + ((x + 0.5 - x_center) ** 2 + (y + 0.5 - y_center) ** 2 < r2)
    i = np.copy(c)

    bool_c0 = np.where(c.ravel() == 0)
    bool_c4 = np.where(c.ravel() == 4)
    i.ravel()[bool_c0] = outer_intensity
    i.ravel()[bool_c4] = inner_intensity

    # compute contour pixels
    delta = 1 / (grid_resolution - 1)

    dx = np.zeros(((np.arange(-0.5, 0.5, delta)).size, (np.arange(-0.5, 0.5, delta)).size), dtype=np.float32).T
    dy = dx.copy()
    dx[:, :] = np.arange(-0.5, 0.5, delta)
    dy_vect = np.arange(-0.5, 0.5, delta)
    dy_vect = np.reshape(dy_vect, (dy_vect.shape[0], 1))
    dy[:, :] = dy_vect

    bool_c04 = np.logical_and(c > 0, c < 4)

    i.ravel()[bool_c04.ravel()] = (outer_intensity +
                                   (inner_intensity - outer_intensity) * circle_grid(
                x.ravel()[(bool_c04.ravel())],
                y.ravel()[(bool_c04.ravel())], r2, x_center, y_center, dx, dy)).reshape((-1,))

    return i


@njit(cache=True)
def circle_vertical_window(x_window_center, y_window_center, x_center, y_center, radius, inner_intensity,
                           outer_intensity, grid_resolution):
    # compute pixels completely outside or inside
    r2 = radius ** 2

    x = np.zeros(((np.arange(x_window_center - 1, x_window_center + 1 + 1)).size,
                  (np.arange(y_window_center - 4, y_window_center + 4 + 1)).size),
                 dtype=np.float64).T
    y = x.copy()
    x[:, :] = np.arange(x_window_center - 1, x_window_center + 1 + 1)
    y_vect = np.arange((y_window_center - 4), (y_window_center + 4 + 1))
    y_vect = np.reshape(y_vect, (y_vect.shape[0], 1))
    y[:, :] = y_vect

    c = ((x - 0.5 - x_center) ** 2 + (y - 0.5 - y_center) ** 2 < r2).astype(np.float64)
    c = c + ((x - 0.5 - x_center) ** 2 + (y + 0.5 - y_center) ** 2 < r2)
    c = c + ((x + 0.5 - x_center) ** 2 + (y - 0.5 - y_center) ** 2 < r2)
    c = c + ((x + 0.5 - x_center) ** 2 + (y + 0.5 - y_center) ** 2 < r2)
    i = np.copy(c)

    bool_c0 = np.where(c.ravel() == 0)
    bool_c4 = np.where(c.ravel() == 4)
    i.ravel()[bool_c0] = outer_intensity
    i.ravel()[bool_c4] = inner_intensity

    # compute contour pixels
    delta = 1 / (grid_resolution - 1)

    dx = np.zeros(((np.arange(-0.5, 0.5, delta)).size, (np.arange(-0.5, 0.5, delta)).size), dtype=np.float32).T
    dy = dx.copy()
    dx[:, :] = np.arange(-0.5, 0.5, delta)
    dy_vect = np.arange(-0.5, 0.5, delta)
    dy_vect = np.reshape(dy_vect, (dy_vect.shape[0], 1))
    dy[:, :] = dy_vect

    bool_c04 = np.logical_and(c > 0, c < 4)
    grid = circle_grid(
                x.ravel()[(bool_c04.ravel())],
                y.ravel()[(bool_c04.ravel())], r2, x_center, y_center, dx, dy)

    i.ravel()[bool_c04.ravel()] = (outer_intensity + (inner_intensity - outer_intensity) * grid).reshape((-1,))

    return i



import numpy as np
from numba import njit

from subpixel_edges.circle import circle_horizontal_window, circle_vertical_window


# @njit(cache=True)
def h_edges(F, G, rows, Gx, Gy, w, edges, order, threshold, x, y, cols,  I, C):
    FF = F.transpose().ravel()
    GG = G.transpose().ravel()

    n_edges = np.shape(edges)[0]

    A = np.zeros((n_edges, 1))
    B = np.zeros((n_edges, 1))
    a = np.zeros((n_edges, 1))
    b = np.zeros((n_edges, 1))
    c = np.zeros((n_edges, 1))
    nx = np.zeros((n_edges, 1))
    ny = np.zeros((n_edges, 1))
    curv = np.zeros((n_edges, 1))


    valid = np.full((n_edges, 1), False)

    pixel_grid_resol = 50

    for k in range(n_edges):
        edge = edges[k]
        m1 = -1
        m2 = 1
        if Gx[edge] * Gy[edge] >= 0:
            m = 1
            l1 = -1
            r2 = 1
            min_l1 = -3
            max_r2 = 3
            l2 = 1
            r1 = -1
            max_l2 = 4
            min_r1 = -4
        else:
            m = -1
            l1 = -1
            r2 = 1
            min_l1 = -4
            max_r2 = 4
            l2 = 1
            r1 = -1
            max_l2 = 3
            min_r1 = -3

        while l1 > min_l1 and np.abs(Gy[edge - rows + l1]) >= np.abs(Gy[edge - rows + l1 - 1]):
            l1 -= 1

        while l2 < max_l2 and np.abs(Gy[edge - rows + l2]) >= np.abs(Gy[edge - rows + l2 + 1]):
            l2 += 1

        while m1 > -4 and np.abs(Gy[edge + m1]) >= np.abs(Gy[edge + m1 - 1]):
            m1 -= 1

        while m2 < 4 and np.abs(Gy[edge + m2]) >= np.abs(Gy[edge + m2 + 1]):
            m2 += 1

        while r1 > min_r1 and np.abs(Gy[edge + rows + r1]) >= np.abs(Gy[edge + rows + r1 - 1]):
            r1 -= 1

        while r2 < max_r2 and np.abs(Gy[edge + rows + r2]) >= np.abs(Gy[edge + rows + r2 + 1]):
            r2 += 1

        window = np.zeros((9, 3))
        window[l1 + 5 - 1:l2 + 5, 0] = 1
        window[m1 + 5 - 1:m2 + 5, 1] = 100
        window[r1 + 5 - 1:r2 + 5, 2] = 1

        if m > 0:
            AA = (GG[edge + m2] + GG[edge + rows + r2]) / 2
            BB = (GG[edge - rows + l1] + GG[edge + m1]) / 2
        else:
            AA = (GG[edge - rows + l2] + GG[edge + m2]) / 2
            BB = (GG[edge + m1] + GG[edge + rows + r1]) / 2

        # search for a second close edge
        u_border = False
        d_border = False

        if m1 > -4:
            partial = np.abs(GG[edge + m1 - 2])
            if partial > np.abs(GG[edge] / 4) and partial > threshold / 2:
                u_border = True

        if m2 < 4:
            partial = np.abs(GG[edge + m2 + 2])
            if partial > np.abs(GG[edge] / 4) and partial > threshold / 2:
                d_border = True

        SL = 0
        SM = 0
        SR = 0
        j = np.int(np.floor((edges[k] - 1) / rows) + 1) 
        i = np.int(edges[k] - rows * (j-1 )) + 1 

        if u_border or d_border:
            rimvt = np.copy(F[i - 5 - 1:i + 5, j - 2 - 1:j + 2])
            if u_border:
                if m > 0:
                    BB = (FF[edge + m1] + FF[edge - rows + l1]) / 2
                    p = 1
                else:
                    BB = (FF[edge + m1] + FF[edge + rows + r1]) / 2
                    p = 0

                if Gy[edge - 2 * rows + l1 + p] * Gy[edge] > 0:
                    ll = l1 + p - 1
                else:
                    ll = l1 + p

                if Gy[edge + 2 * rows + r1 + 1 - p] * Gy[edge] > 0:
                    rr = r1 - p
                else:
                    rr = r1 + 1 - p

                rimvt[0:ll + 6, 0] = BB
                rimvt[0:l1 + 6, 1] = BB
                rimvt[0:m1 + 6, 2] = BB
                rimvt[0:r1 + 6, 3] = BB
                rimvt[0:rr + 6, 4] = BB
                l1 = -3 + m
                m1 = -3
                r1 = -3 - m

            if d_border:
                if m > 0:
                    AA = (FF[edge + m2] + FF[edge + rows + r2]) / 2
                    p = 1
                else:
                    AA = (FF[edge + m2] + FF[edge - rows + l2]) / 2
                    p = 0

                if Gy[edge - 2 * rows + l2 + p - 1] * Gy[edge] > 0:
                    ll = l2 + p
                else:
                    ll = l2 + p - 1

                if Gy[edge + 2 * rows + r2 - p] * Gy[edge] > 0:
                    rr = r2 + 1 - p
                else:
                    rr = r2 - p

                rimvt[ll + 5:11, 0] = AA
                rimvt[l2 + 5:11, 1] = AA
                rimvt[m2 + 5:11, 2] = AA
                rimvt[r2 + 5:11, 3] = AA
                rimvt[rr + 5:11, 4] = AA 

                l2 = 3 + m
                m2 = 3
                r2 = 3 - m

            rimv2 = (rimvt[0:9, 0:3] + rimvt[0:9, 1:4] + rimvt[0:9, 2:5] + \
                     rimvt[1:10, 0:3] + rimvt[1:10, 1:4] + rimvt[1:10, 2:5] + \
                     rimvt[2:11, 0:3] + rimvt[2:11, 1:4] + rimvt[2:11, 2:5]) / 9

            for n in range(l1 + 5 - 1, l2 + 5):
                SL = SL + rimv2[n, 0]
            for n in range(m1 + 5 - 1, m2 + 5):
                SM = SM + rimv2[n, 1]
            for n in range(r1 + 5 - 1, r2 + 5):
                SR = SR + rimv2[n, 2]
        else:
            for n in range(l1, l2 + 1):
                SL = SL + GG[edge - rows + n]
            for n in range(m1, m2 + 1):
                SM = SM + GG[edge + n]
            for n in range(r1, r2 + 1):
                SR = SR + GG[edge + rows + n]

        # compute edge features
        den = 2 * (AA - BB)
        if order == 2:
            if den != 0:
                c[k] = np.divide(SL + SR - 2 * SM + AA * (2 * m2 - l2 - r2) - BB * (2 * m1 - l1 - r1), den)
        else:
            c[k] = 0

        if den != 0:
            a[k] = np.divide(2 * SM - AA * (1 + 2 * m2) - BB * (1 - 2 * m1), den) - w * c[k]
            if np.abs(a[k]).item() > 1:
                valid[k] = False
                continue

            valid[k] = True
            b[k] = np.divide(SR - SL + AA * (l2 - r2) - BB * (l1 - r1), den)
            A[k] = AA
            B[k] = BB
            s = np.sign(AA - BB)
            nx[k] = s / np.sqrt(1 + b[k] ** 2) * b[k]
            ny[k] = s / np.sqrt(1 + b[k] ** 2)
            curv[k] = 2 * c[k] / ((1 + b[k] ** 2) ** 1.5)

            if Gy[edge] < 0:
                curv[k] = -curv[k]

            # generate circle subimage
            if curv[k, 0].item() != 0:
                R = np.abs(1 / curv[k, 0].item())

            RG = float(10000)
            RL = float(4.5)

            if R > RG:
                R = RG
            if R < RL:
                R = RL
            if curv[k].item() > 0:
                s = -1
                inner_intensity = min(AA, BB)
                outer_intensity = max(AA, BB)
            else:
                s = 1
                inner_intensity = max(AA, BB)
                outer_intensity = min(AA, BB)

        center = ([x[edge]+1 + s * R * nx[k], y[edge]+1 - a[k] + s * R * ny[k]])
        subimage = circle_vertical_window(j, i, center[0], center[1], R, inner_intensity, outer_intensity,
                                          pixel_grid_resol)

        # update counter and intensity images
        I[i-4-1:i+4,j-1-1:j+1] = I[i-4-1:i+4,j-1-1:j+1] + window*subimage
        C[i-4-1:i+4,j-1-1:j+1] = C[i-4-1:i+4,j-1-1:j+1] + window

    # remove invalid values
    valid = valid.reshape((-1,))

    edges = edges[valid]
    A = A[valid]
    B = B[valid]
    a = a[valid]
    b = b[valid]
    c = c[valid]
    nx = nx[valid].reshape((-1,))
    ny = ny[valid].reshape((-1,))

    x = x[edges]
    y = y[edges] - a.transpose().ravel()

    curv = curv[valid].reshape((-1,))

    i0 = np.minimum(A, B).reshape((-1,))
    i1 = np.maximum(A, B).reshape((-1,))

    return x, y, edges, nx, ny, i0, i1, curv, I, C, G


# @njit(cache=True)
def v_edges(F, G, rows, Gx, Gy, w, edges, order, threshold, x, y, I, C):
    n_edges = np.shape(edges)[0]

    A = np.zeros((n_edges, 1))
    B = np.zeros((n_edges, 1))
    a = np.zeros((n_edges, 1))
    b = np.zeros((n_edges, 1))
    c = np.zeros((n_edges, 1))
    nx = np.zeros((n_edges, 1))
    ny = np.zeros((n_edges, 1))
    curv = np.zeros((n_edges, 1))

    FF = F.transpose().ravel()
    GG = G.transpose().ravel()

    valid = np.full((n_edges, 1), False)

    pixelGridResol = 50

    # compute all vertical edges
    for k in range(n_edges):
        edge = edges[k]

        # compute window floating limits
        m1 = -1
        m2 = 1

        if Gx[edge] * Gy[edge] >= 0:
            m = 1
            l1 = -1
            r2 = 1
            min_l1 = -3
            max_r2 = 3
            l2 = 1
            r1 = -1
            max_l2 = 4
            min_r1 = -4
        else:
            m = -1
            l1 = -1
            r2 = 1
            min_l1 = -4
            max_r2 = 4
            l2 = 1
            r1 = -1
            max_l2 = 3
            min_r1 = -3

        while l1 > min_l1 and np.abs(Gx[edge - 1 + l1 * rows]) >= np.abs(Gx[edge - 1 + (l1 - 1) * rows]):
            l1 = l1 - 1

        while l2 < max_l2 and np.abs(Gx[edge - 1 + l2 * rows]) >= np.abs(Gx[edge - 1 + (l2 + 1) * rows]):
            l2 = l2 + 1

        while m1 > -4 and np.abs(Gx[edge + m1 * rows]) >= np.abs(Gx[edge + (m1 - 1) * rows]):
            m1 = m1 - 1

        while m2 < 4 and np.abs(Gx[edge + m2 * rows]) >= np.abs(Gx[edge + (m2 + 1) * rows]):
            m2 = m2 + 1

        while r1 > min_r1 and np.abs(Gx[edge + 1 + r1 * rows]) >= np.abs(Gx[edge + 1 + (r1 - 1) * rows]):
            r1 = r1 - 1

        while r2 < max_r2 and np.abs(Gx[edge + 1 + r2 * rows]) >= np.abs(Gx[edge + 1 + (r2 + 1) * rows]):
            r2 = r2 + 1

        window = np.zeros((3, 9))
        window[0, l1 + 5-1:l2 + 5] = 1
        window[1, m1 + 5-1:m2 + 5] = 100
        window[2, r1 + 5-1:r2 + 5] = 1

        # compute intensities
        if m > 0:
            AA = (GG[edge + m2 * rows] + GG[edge + 1 + r2 * rows]) / 2
            BB = (GG[edge - 1 + l1 * rows] + GG[edge + m1 * rows]) / 2
        else:
            AA = (GG[edge - 1 + l2 * rows] + GG[edge + m2 * rows]) / 2
            BB = (GG[edge + m1 * rows] + GG[edge + 1 + r1 * rows]) / 2

        # search for a second close edge
        u_border = False
        d_border = False

        if m1 > -4:
            partial = np.abs(GG[edge + (m1 - 2) * rows])
            if partial > np.abs(GG[edge] / 4) and partial > threshold / 2:
                u_border = True

        if m2 < 4:
            partial = np.abs(GG[edge + (m2 + 2) * rows])
            if partial > np.abs(GG[edge] / 4) and partial > threshold / 2:
                d_border = True

        SL = 0
        SM = 0
        SR = 0
        j = np.int(np.floor((edges[k] - 1) / rows) + 1)
        i = np.int(edges[k] - rows * (j - 1)) + 1

        if u_border or d_border:
            rimvt = np.copy(F[i - 2 - 1:i + 2, j - 5 - 1:j + 5])
            if u_border:
                if m > 0:
                    BB = (FF[edge + m1 * rows] + FF[edge - 1 + l1 * rows]) / 2
                    p = 1
                else:
                    BB = (FF[edge + m1 * rows] + FF[edge + 1 + r1 * rows]) / 2
                    p = 0

                if Gx[edge - 2 + (l1 + p) * rows] * Gx[edge] > 0:
                    ll = l1 + p - 1
                else:
                    ll = l1 + p

                if Gx[edge + 2 + (r1 + 1 - p) * rows] * Gx[edge] > 0:
                    rr = r1 - p
                else:
                    rr = r1 + 1 - p

                rimvt[0, 0:ll + 6] = BB
                rimvt[1, 0:l1 + 6] = BB
                rimvt[2, 0:m1 + 6] = BB
                rimvt[3, 0:r1 + 6] = BB
                rimvt[4, 0:rr + 6] = BB

                l1 = -3 + m
                m1 = -3
                r1 = -3 - m

            if d_border:
                if m > 0:
                    AA = (FF[edge + m2 * rows] + FF[edge + 1 + r2 * rows]) / 2
                    p = 1
                else:
                    AA = (FF[edge + m2 * rows] + FF[edge - 1 + l2 * rows]) / 2
                    p = 0

                if Gx[edge - 2 + (l2 + p - 1) * rows] * Gx[edge] > 0:
                    ll = l2 + p
                else:
                    ll = l2 + p - 1

                if Gx[edge + 2 + (r2 - p) * rows] * Gx[edge] > 0:
                    rr = r2 + 1 - p
                else:
                    rr = r2 - p

                rimvt[0, ll + 6-1:11] = AA
                rimvt[1, l2 + 6-1:11] = AA
                rimvt[2, m2 + 6-1:11] = AA
                rimvt[3, r2 + 6-1:11] = AA
                rimvt[4, rr + 6-1:11] = AA

                l2 = 3 + m
                m2 = 3
                r2 = 3 - m

            rimv2 = (rimvt[0:3, 0:9] + rimvt[1:4, 0:9] + rimvt[2:5, 0:9] +
                     rimvt[0:3, 1:10] + rimvt[1:4, 1:10] + rimvt[2:5, 1:10] +
                     rimvt[0:3, 2:11] + rimvt[1:4, 2:11] + rimvt[2:5, 2:11]) / 9

            for n in range(l1 + 5 - 1, l2 + 5):
                SL = SL + rimv2[0, n]
            for n in range(m1 + 5 - 1, m2 + 5):
                SM = SM + rimv2[1, n]
            for n in range(r1 + 5 - 1, r2 + 5):
                SR = SR + rimv2[2, n]
        else:

            for n in range(l1, l2 + 1):
                SL = SL + GG[edge - 1 + n * rows]
            for n in range(m1, m2 + 1):
                SM = SM + GG[edge + n * rows]
            for n in range(r1, r2 + 1):
                SR = SR + GG[edge + 1 + n * rows]

        # compute edge features
        den = 2 * (AA - BB)
        if order == 2:
            if den != 0:
                c[k] = np.divide((SL + SR - 2 * SM + AA * (2 * m2 - l2 - r2) - BB * (2 * m1 - l1 - r1)), den)
        else:
            c[k] = 0

        if den != 0:
            a[k] = np.divide((2 * SM - AA * (1 + 2 * m2) - BB * (1 - 2 * m1)), den) - w * c[k]
            if np.abs(a[k].item()) > 1:
                valid[k] = False
                continue

            valid[k] = True
            b[k] = np.divide((SR - SL + AA * (l2 - r2) - BB * (l1 - r1)), den)
            A[k] = AA
            B[k] = BB

            s = np.sign(AA - BB)
            nx[k] = s / np.sqrt(1 + b[k] ** 2)
            ny[k] = s / np.sqrt(1 + b[k] ** 2) * b[k]
            curv[k] = 2 * c[k] / ((1 + b[k] ** 2) ** 1.5)
            if Gx[edge] < 0:
                curv[k] = -curv[k]

            # generate circle subimage
            if curv[k].item() != 0:
                R = np.abs(1 / curv[k].item())
            if R > 1e4:
                R = 1e4

            if R < 4.5:
                R = 4.5

            if curv[k].item() > 0:
                s = -1
                inner_intensity = min(AA, BB)
                outer_intensity = max(AA, BB)
            else:
                s = 1
                inner_intensity = max(AA, BB)
                outer_intensity = min(AA, BB)

        center = [x[edge] + 1 - a[k] + s * R * nx[k], y[edge] + 1 + s * R * ny[k]]
        subimage = circle_horizontal_window(j, i, center[0], center[1],
                                                R, inner_intensity, outer_intensity, pixelGridResol)

        # update counter and intensity images
        I[i-1-1:i+1,j-4-1:j+4] = I[i-1-1:i+1,j-4-1:j+4] + window*subimage
        C[i-1-1:i+1,j-4-1:j+4] = C[i-1-1:i+1,j-4-1:j+4] + window

    edges = edges[valid.reshape((-1,))]
    A = A[valid.reshape((-1,))]
    B = B[valid.reshape((-1,))]
    a = a[valid.reshape((-1,))]
    b = b[valid.reshape((-1,))]
    c = c[valid.reshape((-1,))]
    nx = nx[valid.reshape((-1,))].reshape((-1,))
    ny = ny[valid.reshape((-1,))].reshape((-1,))

    x = x[edges] - a.transpose().ravel()
    y = y[edges]

    curv = curv[valid.reshape((-1,))].reshape((-1,))

    i0 = np.minimum(A, B).reshape((-1,))
    i1 = np.maximum(A, B).reshape((-1,))

    return x, y, edges, nx, ny, i0, i1, curv, I, C, G
